with Messages, Ada.Text_IO, Reps, Reps.Ops, Reps.IO, Lattice, Lattice.Ops;
use Messages, Ada.Text_IO, Reps, Reps.Ops, Reps.IO, Lattice, Lattice.Ops;

procedure Stats is

  LP:   constant Integer := GetArg(1);  -- layer period: set 2 for AB, 3 for ABC
  NP:   constant Integer  := GetArg(2); -- size of polymers
  NRes: constant Integer  := GetArg(3); -- const/NRes is width of Gaussian

  Steps: constant Integer := 1000;

  N1: constant Positive := 12;  -- lattice patch depth:  -N1 .. N1
  N2: constant Positive := 12;  -- lattice patch width:  -N2 .. N1
  N3: constant Positive :=  8;  -- lattice patch height: -N3 .. N3

  Eps: constant Rep := 1.0E-7;

  function "**"(S: Rep; I: Integer) return Rep is
  begin
    if I>0 then
      declare
        J: Integer := I;
        X,Y: Rep;
      begin
        X := S;
        if (J rem 2)=0 then
          Y := One;
        else
          Y := X;
        end if;
        J := J/2;
        while J>0 loop
          X := Sqr(X);
          if (J rem 2)>0 then
            Y := X*Y;
          end if;
          J := J/2;
        end loop;
        return Y;
      end;
    elsif I<0 then
      return (One/S)**(-I);
    else
      return One;
    end if;
  end "**";

  function Moment(P: PolysPtr; I: Integer) return Rep is
    N: Integer := 0;
    Y: Rep := Zero;
  begin
    for K in P'Range loop
      N := N+P(K).M;
      Y := Y+Rep(P(K).M)*P(K).R**I;
    end loop;
    return Y/Rep(N);
  end Moment;

  function NumPoints(P: PolysPtr) return Integer is
    N: Integer := 0;
  begin
    for K in P'Range loop
      N := N+P(K).M;          -- add multiplicities
    end loop;
    return N;
  end NumPoints;

  function Lowest(P: PolysPtr) return Rep is
    E: Rep := Rep'Last;
  begin
    for K in P'Range loop E := RMin(E,P(K).R); end loop;
    return E;
  end Lowest;

  function Highest(P: PolysPtr) return Rep is
    E: Rep := -Rep'Last;
  begin
    for K in P'Range loop E := RMax(E,P(K).R); end loop;
    return E;
  end Highest;

  function Gauss(S,X: Rep) return Rep is
    Pi: constant Rep := 3.141592653589793;
  begin
    return Exp(-Sqr(X/S)/Two)/(S*Sqrt(Two*Pi));
  end Gauss;

  function Gconv(P: PolysPtr; S,X: Rep) return Rep is
    N: Integer := 0;
    R: Rep := Zero;
  begin
    for K in P'Range loop
      N := N+P(K).M;
      R := R+Rep(P(K).M)*Gauss(S,X-P(K).R);
    end loop;
    return R/Rep(N);
  end Gconv;

  function Hat(S,X: Rep) return Rep is
    R: Rep := Zero;
  begin
    if abs(X)<S then return Half/S; else return Zero; end if;
  end Hat;

  function Hconv(P: PolysPtr; S,X: Rep) return Rep is
    N: INteger := 0;
    R: Rep := Zero;
  begin
    for K in P'Range loop
      N := N+P(K).M;
      R := R+Rep(P(K).M)*Hat(S,X-P(K).R);
    end loop;
    return R/Rep(N);
  end Hconv;

  procedure Show(X,Y: in Rep) is
  begin
    Txt_Put(Current_Output,X);
    Put("  ");
    Txt_Put(Current_Output,Y);
    New_Line;
  end Show;

  procedure Moments(P: in PolysPtr) is
    E1,E2,E3,E4,T1,T2,T3,T4: Rep;
  begin
    E1 := Moment(P,1);
    E2 := Moment(P,2);
    E3 := Moment(P,3);
    E4 := Moment(P,4);
    T1 := E1;
    T2 := E2-T1**2;
    T3 := E3-Rep(3)*T1*T2-T1*T1*T1;
    T4 := E4-Rep(4)*T1*T3-Rep(3)*T2*T2-T1*T1*T1*T1;
    Show("1st Moment:",E1);
    Show("Variance:  ",Sqrt(T2));
    Show("1st Moment / Variance:  ",E1/Sqrt(T2));

--    Show("2nd Moment:",E2);
--    Show("3rd Moment:",E3);
--    Show("4th Moment:",E4);
--    Show("2nd Trunc: ",T2);
--    Show("3rd Trunc: ",T3);
--    Show("4th Trunc: ",T4);

  end Moments;

  Elow,Ehigh,SG,DX,X,Y: Rep;

  PEP: PExpPtr;
  PP: PolysPtr;

begin
  Verbose := 0;
  InitPatch(N1,N2,N3,LP);
  PEP := Entropies(NP);
  PP := PEP(NP);

  Elow  := -0.00030; --- Lowest(PP);
  EHigh :=  0.00035; --- Highest(PP);

  SG := (EHigh-ELow)/Rep(NRes);
  DX := (EHigh-ELow)/Rep(Steps);

  X := Elow-Eps;

--  Moments(PP);

-- return;

  loop
    Y := Gconv(PP,SG,X);
    Show(X,Y);
    exit when X>Ehigh+Eps;
    X := X+DX;
  end loop;

end Stats;
